#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <stdio.h>
#include "msi.h"

#define MAX 16 /* maximum remote slaves supported */
#define LEN 80 /* maximum hostname length allowed */

#define FALSE 0
#define TRUE 1

/* Boolean flags */
int Running = TRUE;
int Alarm = FALSE;

unsigned char bcaddr[4] = {0xC0, 0x40, 0x09, 0xFF}; /* 192.64.9.255 */

main(void)
/*
**   master.c - Master feeds queries to Slave services 
*/
{
   int len = sizeof(struct sockaddr_in);
   struct sockaddr_in broadcast;
   struct sockaddr_in registers;
   struct sockaddr_in querylink;
   struct sockaddr_in slavelink;
   struct hostent *hp;
#ifdef TIMING
 struct timeval workval, phase1[MAX]; /* phase 1 happens in Parent */
 struct timeval ph01wrk;    /* choose an IDLE subtask from the pool */
 struct timeval ph02wrk;    /* subtask sends incoming query to remote */
 struct timeval ph03wrk;    /* calculated Master-to-Slave transit time */
 struct timeval ph04wrk;    /* Slave gets query and extracts request */
 struct timeval ph05wrk;    /* Slave retrieves line of text from file */
 struct timeval ph06wrk;    /* Slave builds response and sends it back */
 struct timeval ph07wrk;    /* calculated Slave-to-Master transit time */
 struct timeval ph08wrk;    /* subtask responds to query and sets IDLE */
 double tot, avg;
 int pkts=0;                /* incoming packet counter */
 FILE *tfp;
#endif
   void sig_inter();
   void sig_alarm();

   /*
   **   Shared Memory Space
   */
   struct area {            /* flag for every subtask */
      int flag[MAX];        /* 0:idle, 1:good, -:exit */
#ifdef TIMING
 int count[MAX];            /* number of queries handled */
 struct timeval phase2[MAX];
 struct timeval phase3[MAX];
 struct timeval phase4[MAX];
 struct timeval phase5[MAX];
 struct timeval phase6[MAX];
 struct timeval phase7[MAX];
 struct timeval phase8[MAX];
#endif
   } area, *aptr;

   int flag = (IPC_CREAT | IPC_EXCL | 0660);
   int size = sizeof(struct area);
   key_t key = 0x01234567;  /* shared memory area key */
   int shmid;               /* shared memory area id */
   int acc;                 /* accept file descriptor */
   int cli;                 /* query file descriptor */
   int pid;                 /* performance monitor */
   int b_fd;                /* broadcast file desc */
   int r_fd;                /* registration file desc */
   int s_fd;                /* slave link file desc */
   int ident;               /* subtask id and offset */
   int opt=1;
   fd_set fds;              /* select list for queries */
   int count=0;             /* registered slave count */
   int pids[MAX];           /* subtask process idents */
   char hostname[LEN];      /* host where Master runs */
   char hosts[MAX][LEN];    /* registered slave hosts */
   char buf[QUERYSIZE], *p; /* client query/response */
 
   /* init signal handlers */
   signal(SIGINT, sig_inter);
   signal(SIGALRM, sig_alarm);

   /* establish a UDP broadcast socket */
   b_fd = socket(AF_INET, SOCK_DGRAM, 0);
   memset((char *) &broadcast, 0, len);
   broadcast.sin_family = AF_INET;
   broadcast.sin_port = htons(BROADCAST);
   memcpy(&broadcast.sin_addr, bcaddr, 4);
   setsockopt(b_fd, SOL_SOCKET, SO_BROADCAST,
      (char *) &opt, sizeof(opt));

   /* establish a registrations socket */
   r_fd = socket(AF_INET, SOCK_DGRAM, 0);
   memset((char *) &registers, 0, len);
   registers.sin_addr.s_addr = htonl(INADDR_ANY);
   registers.sin_port = htons(REGISTERS);
   registers.sin_family = AF_INET;
   bind(r_fd, (struct sockaddr *) &registers, len);

   /* start up all remote slave services */
   gethostname(hostname, sizeof(hostname));
   if((p=strchr(hostname, '.')) != NULL)
      *p = '\0'; /* trunc .domainname */
   sendto(b_fd, hostname, strlen(hostname)+1, 0,
      (struct sockaddr *) &broadcast, len);
   close(b_fd);

   /*
   **   Display registration response info
   */
   while(1) /* alarm signal causes exit */
   {
      alarm(1); /* set a 1-second alarm */
      buf[0] = '\0'; /* clear input buf */
      /* wait for registration or alarm */
      recvfrom(r_fd, buf, sizeof(buf), 0,
        (struct sockaddr *) &registers, &len);
      if(Alarm)
         break;
      alarm(0); /* cancel alarm timer */
      if((p=strchr(buf, '.')) != NULL)
         *p = '\0'; /* trunc .domain */
      printf("\tSlave on host '%s' registered\n", buf);
      strcpy(hosts[count++], buf);
   }
   close(r_fd);
   printf("\tRegistered %d remote slave task(s)", count);

   /*
   **   ERROR if unsupported #slaves
   */
   if((count <= 0) || (count > MAX))
   {
      printf(" - Unsupported!\n");
      exit(-1);
   }
   printf(".\n\n");

   /* start "perform" */
   if((pid=fork()) == 0)
   {
      execl("/home/chief/bin/perform", "perform", "cardinal", NULL);
      exit(-1); /* a successful execl never returns to the caller */
   }
 
   /*
   **   Shared memory area for subtask pool
   */
   size *= count;
   shmid = shmget(key, size, flag);
   aptr = (struct area *) shmat(shmid, 0, 0);

   /*
   **   TCP socket for client's queries
   */
   acc = socket(AF_INET, SOCK_STREAM, 0);
   querylink.sin_family = AF_INET;
   querylink.sin_port = htons(QUERYPORT);
   querylink.sin_addr.s_addr = htonl(INADDR_ANY);
   bind(acc, (struct sockaddr *) &querylink, len);
   setsockopt(acc, SOL_SOCKET, SO_REUSEADDR,
      (char *) &opt, sizeof(opt));
   listen(acc, 5);

   /*
   **   Pre-establish a subtask pool
   */
   for(ident=0; ident<count; ident++)
   {
      aptr->flag[ident] = 0; /* idle */
#ifdef TIMING
 aptr->count[ident] = 0; /* zero cumulative statistics */
 phase1[ident].tv_sec = 0L;
 phase1[ident].tv_usec = 0L;
 aptr->phase2[ident].tv_sec = 0L;
 aptr->phase2[ident].tv_usec = 0L;
 aptr->phase3[ident].tv_sec = 0L;
 aptr->phase3[ident].tv_usec = 0L;
 aptr->phase4[ident].tv_sec = 0L;
 aptr->phase4[ident].tv_usec = 0L;
 aptr->phase5[ident].tv_sec = 0L;
 aptr->phase5[ident].tv_usec = 0L;
 aptr->phase6[ident].tv_sec = 0L;
 aptr->phase6[ident].tv_usec = 0L;
 aptr->phase7[ident].tv_sec = 0L;
 aptr->phase7[ident].tv_usec = 0L;
 aptr->phase8[ident].tv_sec = 0L;
 aptr->phase8[ident].tv_usec = 0L;
#endif

      if((pids[ident]=fork()) == 0)                   /* SUBTASK */
      {
         /* nullify ^C handler */
         signal(SIGINT, SIG_IGN);

         /* attach the shared memory flags array */
         aptr = (struct area *) shmat(shmid, 0, 0);
         /* notice: shmid is still set correctly */

         /*
         **   Establish slave comm link socket
         */
         s_fd = socket(AF_INET, SOCK_DGRAM, 0);
         memset((char *) &slavelink, 0, len);
         hp = gethostbyname(hosts[ident]);
         memcpy(&slavelink.sin_addr, hp->h_addr, hp->h_length);
         slavelink.sin_port = htons(SLAVEPORT);
         slavelink.sin_family = AF_INET;
         bind(s_fd, (struct sockaddr *) &slavelink, len);

         /*
         **   Poll on our flag until it becomes non-zero
         */
         while(1)
         {
            if(aptr->flag[ident] == 0) /* do we stay idle? */
            {
               usleep(1);   /* release processor */
               continue;    /* stay in poll loop */
            }
            else if(aptr->flag[ident] < 0) /* exit request? */
            {
               /* signal remote slave to exit */
               strcpy(packet.buffer, "Terminate 0");
               sendto(s_fd, &packet, sizeof(packet),
                  0, (struct sockaddr *) &slavelink, len);
               break; /* SUBTASKs normal exit */
            }
            else /* action flag must be positive */
            {
               cli = accept(acc, (struct sockaddr *) &querylink, &len);
#ifdef TIMING
 gettimeofday(&ph02wrk, NULL); /* start Ph2 */
#endif
               /* recv incoming query from client */
               packet.buffer[0] = 0; /* clear buf */
               recv(cli, packet.buffer, QUERYSIZE, 0);

               /* give packet to slave task for processing */
#ifdef TIMING
 gettimeofday(&workval, NULL); /* end Ph2 */
 workval.tv_sec -= ph02wrk.tv_sec;
 workval.tv_usec -= ph02wrk.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 aptr->phase2[ident].tv_sec += workval.tv_sec;
 aptr->phase2[ident].tv_usec += workval.tv_usec;
 /* start transmission time */
 gettimeofday(&ph03wrk, NULL); /* start Ph3 through Ph7 */
#endif
               sendto(s_fd, &packet, sizeof(struct packet),
                  0, (struct sockaddr *) &slavelink, len);
               packet.buffer[0] = 0; /* empty the buffer */
               recvfrom(s_fd, &packet, sizeof(struct packet),
                  0, (struct sockaddr *) &slavelink, &len);
#ifdef TIMING
 gettimeofday(&workval, NULL); /* end Ph3 through Ph7 */
 /* calulate send+slave+recv time */
 workval.tv_sec -= ph03wrk.tv_sec;
 workval.tv_usec -= ph03wrk.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 /* subtract off slave processing time */
 workval.tv_sec -= packet.remote.tv_sec;
 workval.tv_usec -= packet.remote.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 /* divide round-trip time by two */
 ph03wrk.tv_sec = workval.tv_sec / 2;
 if((ph03wrk.tv_usec * 2) < workval.tv_sec)
    workval.tv_usec += 500000L;
 aptr->phase3[ident].tv_sec += workval.tv_sec / 2;
 aptr->phase3[ident].tv_usec += workval.tv_usec / 2;
 /* put the other half into Ph7 */
 aptr->phase7[ident].tv_sec += workval.tv_sec / 2;
 aptr->phase7[ident].tv_usec += workval.tv_usec / 2;
 /* capture Slave's timing statistics from packet */
 aptr->phase4[ident].tv_sec += packet.phase4.tv_sec;
 aptr->phase4[ident].tv_usec += packet.phase4.tv_usec;
 aptr->phase5[ident].tv_sec += packet.phase5.tv_sec;
 aptr->phase5[ident].tv_usec += packet.phase5.tv_usec;
 aptr->phase6[ident].tv_sec += packet.phase6.tv_sec;
 aptr->phase6[ident].tv_usec += packet.phase6.tv_usec;
 gettimeofday(&ph08wrk, NULL); /* start Ph8 */
#endif
               /* result back to client */
               strcpy(buf, packet.buffer);
               send(cli, buf, strlen(buf)+1, 0);
               close(cli);

               /* check for exit request before continuing */
               if(aptr->flag[ident] < 0) /* exit request? */
               {
                  sendto(s_fd, "Terminate 0", 12, 0,
                     (struct sockaddr *) &slavelink, len);
                  break;
               }
               aptr->flag[ident] = 0; /* no: flag as idle */
#ifdef TIMING
 gettimeofday(&workval, NULL); /* end Ph8 */
 workval.tv_sec -= ph08wrk.tv_sec;
 workval.tv_usec -= ph08wrk.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 aptr->phase8[ident].tv_sec += workval.tv_sec;
 aptr->phase8[ident].tv_usec += workval.tv_usec;

 aptr->count[ident]++; /* increment counter */
#endif
            }
         }
         close(s_fd);
         exit(0);                                 /* end SUBTASK */
      }
   } /* for count */

   /*
   **   Pass incoming queries to an available subtask
   */
   ident = 0;
   while(Running)                                 /* start PARENT */
   {
      FD_ZERO(&fds);
      FD_SET(acc, &fds);
      /* block until client is ready for connect */
      if(select(acc+1, &fds, NULL, NULL, NULL) < 0)
         break;
      /*
      **   Find an idle subtask to process a query
      */
#ifdef TIMING
 gettimeofday(&ph01wrk, NULL); /* start Ph1 */
#endif
      for(ident=0; ident<count; ident++)
      {
         if(aptr->flag[ident] == 0) /* available? */
         {
            aptr->flag[ident] = 1; /* work to do! */
            break;
         }
      }
      if(ident == count) /* no available subtasks */
      {
         cli = accept(acc, (struct sockaddr *) &querylink, &len);
         send(cli, "DROPPED", 8, 0);
         close(cli);
      }
#ifdef TIMING
 /* Ph1 is the only Parent processing time */
 gettimeofday(&workval, NULL); /* end Ph1 */
 workval.tv_sec -= ph01wrk.tv_sec;
 workval.tv_usec -= ph01wrk.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 if(ident < count) /* not DROPPED queries */
 {
    phase1[ident].tv_sec += workval.tv_sec;
    phase1[ident].tv_usec += workval.tv_usec;
 }
#endif
      usleep(1);
   } /* while Running */

   /*
   **   Tell each subtask to exit gracefully
   */
   for(ident=0; ident<count; ident++)
   {
      aptr->flag[ident] = -1;
      waitpid(pids[ident]);
   }
   kill(pid, 9);
   waitpid(pid);

#ifdef TIMING
   tfp = fopen("master.txt", "w");
   /* print final timing statistics */
   fprintf(tfp, "\n\tSLAVENAME   PH   NUM       TOTAL     AVERAGE\n");
   fprintf(tfp, "\t---------   --   ---       -----     -------\n");
   ph01wrk.tv_sec = 0L;
   ph01wrk.tv_usec = 0L;
   ph02wrk.tv_sec = 0L;
   ph02wrk.tv_usec = 0L;
   ph03wrk.tv_sec = 0L;
   ph03wrk.tv_usec = 0L;
   ph04wrk.tv_sec = 0L;
   ph04wrk.tv_usec = 0L;
   ph05wrk.tv_sec = 0L;
   ph05wrk.tv_usec = 0L;
   ph05wrk.tv_sec = 0L;
   ph05wrk.tv_usec = 0L;
   ph06wrk.tv_sec = 0L;
   ph06wrk.tv_usec = 0L;
   ph07wrk.tv_sec = 0L;
   ph07wrk.tv_usec = 0L;
   ph08wrk.tv_sec = 0L;
   ph08wrk.tv_usec = 0L;
   for(ident=0; ident<count; ident++)
   {
      pkts += aptr->count[ident];
      tot = ((double) phase1[ident].tv_sec +
            ((double) phase1[ident].tv_usec / 1000000.0));
      ph01wrk.tv_sec += phase1[ident].tv_sec;
      ph01wrk.tv_usec += phase1[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t %8s  %3d     1  %10.6lf  %10.6lf\n", hosts[ident],
             aptr->count[ident], tot, avg);
      tot = ((double) aptr->phase2[ident].tv_sec +
            ((double) aptr->phase2[ident].tv_usec / 1000000.0));
      ph02wrk.tv_sec += aptr->phase2[ident].tv_sec;
      ph02wrk.tv_usec += aptr->phase2[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   2  %10.6lf  %10.6lf\n", tot, avg);
      tot = ((double) aptr->phase3[ident].tv_sec +
            ((double) aptr->phase3[ident].tv_usec / 1000000.0));
      ph03wrk.tv_sec += aptr->phase3[ident].tv_sec;
      ph03wrk.tv_usec += aptr->phase3[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   3  %10.6lf  %10.6lf\n", tot, avg);
      tot = ((double) aptr->phase4[ident].tv_sec +
            ((double) aptr->phase4[ident].tv_usec / 1000000.0));
      ph04wrk.tv_sec += aptr->phase4[ident].tv_sec;
      ph04wrk.tv_usec += aptr->phase4[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   4  %10.6lf  %10.6lf\n", tot, avg);
      tot = ((double) aptr->phase5[ident].tv_sec +
            ((double) aptr->phase5[ident].tv_usec / 1000000.0));
      ph05wrk.tv_sec += aptr->phase5[ident].tv_sec;
      ph05wrk.tv_usec += aptr->phase5[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   5  %10.6lf  %10.6lf\n", tot, avg);
      tot = ((double) aptr->phase6[ident].tv_sec +
            ((double) aptr->phase6[ident].tv_usec / 1000000.0));
      ph06wrk.tv_sec += aptr->phase6[ident].tv_sec;
      ph06wrk.tv_usec += aptr->phase6[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   6  %10.6lf  %10.6lf\n", tot, avg);
      tot = ((double) aptr->phase7[ident].tv_sec +
            ((double) aptr->phase7[ident].tv_usec / 1000000.0));
      ph07wrk.tv_sec += aptr->phase7[ident].tv_sec;
      ph07wrk.tv_usec += aptr->phase7[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   7  %10.6lf  %10.6lf\n", tot, avg);
      tot = ((double) aptr->phase8[ident].tv_sec +
            ((double) aptr->phase8[ident].tv_usec / 1000000.0));
      ph08wrk.tv_sec += aptr->phase8[ident].tv_sec;
      ph08wrk.tv_usec += aptr->phase8[ident].tv_usec;
      if(aptr->count[ident] > 0)
         avg = tot / (double) aptr->count[ident];
      else
         avg = 0.0;
      fprintf(tfp, "\t                   8  %10.6lf  %10.6lf\n", tot, avg);
   }
   tot = ((double) ph01wrk.tv_sec +
         ((double) ph01wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t  OVERALL    1   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph02wrk.tv_sec +
         ((double) ph02wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             2   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph03wrk.tv_sec +
         ((double) ph03wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             3   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph04wrk.tv_sec +
         ((double) ph04wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             4   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph05wrk.tv_sec +
         ((double) ph05wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             5   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph06wrk.tv_sec +
         ((double) ph06wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             6   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph07wrk.tv_sec +
         ((double) ph07wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             7   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   tot = ((double) ph08wrk.tv_sec +
         ((double) ph08wrk.tv_usec / 1000000.0));
   avg = tot / (double) pkts;
   fprintf(tfp, "\t             8   %3d  %10.6lf  %10.6lf\n", pkts, tot, avg);
   fclose(tfp);
#endif

   /* remove unused shared memory area and exit */
   shmctl(shmid, IPC_RMID, (struct shmid_ds *) 0);
   printf("\tDone.\n");
   return(0);
}                                                  /* end PARENT */

/**************/
/*** alarms ***/
/**************/
void sig_alarm()
{
   /* reestablish handler */
   signal(SIGALRM, sig_alarm);
   Alarm = TRUE; /* set flag */
}

/**************/
/*** ^C key ***/
/**************/
void sig_inter()
{
   Running = FALSE; /* set flag */
   return;
}
