#include <netinet/in.h>
#include <sys/socket.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include "cpm.h"

main(void)
{
   /*
   **   monitor.c - Centralized Performance Monitoring
   */
   int cp[4], rm[4]; /* central processor, real memory */
   int ps[4], pr[4]; /* packets sent, packets received */
   int dr[4], dw[4]; /* disk reads, disk writes */
   union {
      unsigned long addr;
      char bytes[4];
   } u;
   struct sockaddr_in soc;
   struct cprStats stats;
   int len=sizeof(soc);
   struct hostent *hp;
   int fd, ii, jj;
   int netLoad;
   char buf[256];

   /* establish and initialize UDP socket struct */
   fd = socket(AF_INET, SOCK_DGRAM, 0);
   memset((char *) &soc, 0, sizeof(soc));
   soc.sin_addr.s_addr = htonl(INADDR_ANY);
   soc.sin_port = htons(PORT);
   soc.sin_family = AF_INET;
   bind(fd, (struct sockaddr *) &soc, len);

   /* initialize stats */
   for(ii=0; ii<4; ii++)
   {
      cp[ii] = 0;
      rm[ii] = 0;
      ps[ii] = 0;
      pr[ii] = 0;
      dr[ii] = 0;
      dw[ii] = 0;
   }

   /*
   **   Centralized Performance Reporting Loop
   */
   while(1)
   {
      for(ii=0; ii<4; ii++)
      {
         recvfrom(fd, &stats, sizeof(stats), 0,
           (struct sockaddr *) &soc, &len);
         /* index is last IP addr digit-2 */
         memcpy(&u.addr, &soc.sin_addr, 4);
         jj = (unsigned int) u.bytes[3];
         cp[jj-2] = stats.cp;
         rm[jj-2] = stats.rm;
         ps[jj-2] = stats.ps;
         pr[jj-2] = stats.pr;
         dr[jj-2] = stats.dr;
         dw[jj-2] = stats.dw;
      }

      /*
      **   Calculate network load
      */
      netLoad = 0;
      for(ii=0; ii<4; ii++)
         netLoad += (ps[ii] + pr[ii]);
      netLoad /= 2; /* average of each node's number sent and received */

      /*
      **   Draw a network diagram showing cluster performance statistics
      */
      system("clear");
      printf("\n      NEAR REAL-TIME CLUSTER PERFORMANCE STATISTICS     \n");
      printf("                                                        \n");
      printf("                         10Base2                        \n");
      printf(" +----ALPHA-----+           |           +-----BETA-----+\n");
      printf(" |   Cpu    Mem |           |           |   Cpu    Mem |\n");
      printf(" |  %3d%%   %3d%% |Rcvd %-5d | %5d Rcvd|  %3d%%   %3d%% |\n",
         cp[0], rm[0], pr[0], pr[1], cp[1], rm[1]);
      printf(" |   Rio    Wio +-----------+-----------+   Rio    Wio |\n");
      printf(" |%6d %6d |Sent %-5d | %5d Sent|%6d %6d |\n",
         dr[0], dw[0], ps[0], ps[1], dr[1], dw[1]);
      printf(" +---10.0.0.1---+           |           +---10.0.0.2---+\n");
      printf("                            |                           \n");
      printf(" +----GAMMA-----+           |           +----DELTA-----+\n");
      printf(" |   Cpu    Mem |           |           |   Cpu    Mem |\n");
      printf(" |  %3d%%   %3d%% |Rcvd %-5d | %5d Rcvd|  %3d%%   %3d%% |\n",
         cp[2], rm[2], pr[2], pr[3], cp[3], rm[3]);
      printf(" |   Rio    Wio +-----------+-----------+   Rio    Wio |\n");
      printf(" |%6d %6d |Sent %-5d | %5d Sent|%6d %6d |\n",
         dr[2], dw[2], ps[2], ps[3], dr[3], dw[3]);
      printf(" +---10.0.0.3---+           |           +---10.0.0.4---+\n");
      printf("                        chaos.org\n\n");
      printf("               - Overall Network Loading -\n");
      printf("                %9d Pkts/sec\n", netLoad);
      printf("\n");
   }
}
