#include <netinet/in.h>
#include <sys/socket.h>
#include <string.h>
#include <netdb.h>
#include <stdio.h>
#include "cpm.h"

main(int argc, char *argv[])
{
/*
**   perform.c - Performance gatherer sends node utilization to monitor.c
*/
   struct sockaddr_in soc;
   struct cprStats stats;
   struct hostent *hp;
   long usec0, usec1; /* user */
   long nsec0, nsec1; /* nice */
   long ssec0, ssec1; /* system */
   long isec0, isec1; /* idle */
   long rpkts0, rpkts1; /* rcvd */
   long spkts0, spkts1; /* sent */
   long tmem, umem;
   long rio0, rio1;
   long wio0, wio1;
   FILE *fptr;
   float fcpu;
   float fmem;
   int ii, jj;
   int icpu;
   int imem;
   int sum;
   int fd;
   char local[16];
   char remote[16];
   char buf[256], *p;

   if(argc != 2)
   {
      printf("\n\tUsage: %s <hostname>\n\n", argv[0]);
      exit(-1);
   }
   /* get remote hostname */
   strcpy(remote, argv[1]);
   /* and the local hostname, also */
   gethostname(local, sizeof(local));
   if((p = strchr(local, '.')) != NULL)
      *p = '\0'; /* trunc domain.nam */

   /*
   **   Central Performance Reporting socket 
   */
   fd = socket(AF_INET, SOCK_DGRAM, 0);
   memset((char *) &soc, 0, sizeof(soc));
   hp = gethostbyname(remote);
   memcpy(&soc.sin_addr, hp->h_addr, hp->h_length);
   soc.sin_port = htons(PORT);
   soc.sin_family = AF_INET;

   /*
   **   Initial CPU utilization
   */
   fptr = fopen("/proc/stat", "r");
   fgets(buf, sizeof(buf), fptr);
   buf[strlen(buf)-1] = 0;
   sscanf(buf, "%*s %ld %ld %ld %ld",
      &usec0, &nsec0, &ssec0, &isec0);
   /*
   **   Initial disk IO rates (also in /proc/stat)
   */
   while((fgets(buf, sizeof(buf), fptr)) != NULL)
   {
      if(strncmp(buf, "disk_", 5) == 0)
         break;
   }
   sscanf(buf, "%*s %ld", &rio0);
   fgets(buf, sizeof(buf), fptr);
   sscanf(buf, "%*s %ld", &wio0);
   fclose(fptr);

   /*
   **   Initial network I/O rates
   */
   fptr = fopen("/proc/net/dev", "r");
   /* skip column header strings in net/dev */
   fgets(buf, sizeof(buf), fptr); /* skip */
   fgets(buf, sizeof(buf), fptr); /* skip */
   fgets(buf, sizeof(buf), fptr); /* skip */
   fgets(buf, sizeof(buf), fptr); /* eth0 */
   buf[strlen(buf)-1] = 0; /* over '\n' */
   buf[6] = ' ';           /* over  ':' */
   sscanf(buf, "%*s %ld %*d %*d %*d %*d %ld", &rpkts0, &spkts0);
   fclose(fptr);

   /*
   **   Statistics-Gathering Loop 
   */
   while(1)
   {
      /*
      **   Real MEM utilization
      */
      fptr = fopen("/proc/meminfo", "r");
      fgets(buf, sizeof(buf), fptr); /* skip hdr */
      fgets(buf, sizeof(buf), fptr);
      buf[strlen(buf)-1] = 0;
      sscanf(buf, "%*s %ld %ld", &tmem, &umem);
      fmem = (float) umem / (float) tmem;
      stats.rm = (int) (100 * (fmem + .005));
      fclose(fptr);

      /*
      **   Subsequent CPU utilization
      */
      fptr = fopen("/proc/stat", "r");
      fgets(buf, sizeof(buf), fptr);
      buf[strlen(buf)-1] = 0;
      sscanf(buf, "%*s %ld %ld %ld %ld",
         &usec1, &nsec1, &ssec1, &isec1);
      /* calculate %CPU utilization */
      usec1 -= usec0; /* 1-second user CPU counter */
      nsec1 -= nsec0; /*          nice             */
      ssec1 -= ssec0; /*          system           */
      isec1 -= isec0; /*          idle             */
      fcpu = (float) (usec1 + nsec1 + ssec1)
           / (float) (usec1 + nsec1 + ssec1 + isec1);
      stats.cp = (int) (100 * (fcpu + .005));

      /*
      **   Subsequent disk I/O rates (also in /proc/stat)
      */
      while((fgets(buf, sizeof(buf), fptr)) != NULL)
      {
         if(strncmp(buf, "disk_rio", 8) == 0) /* skip */
            break;
      }
      sscanf(buf, "%*s %ld", &rio1);
      /* next string after disk_rio is disk_wio */
      fgets(buf, sizeof(buf), fptr);
      sscanf(buf, "%*s %ld", &wio1);
      fclose(fptr);
      /* calculate disk I/O rates */
      stats.dr = (rio1 - rio0) / 5;
      stats.dw = (wio1 - wio0) / 5;
      fclose(fptr);

      /*
      **   Subsequent network I/O rates
      */
      fptr = fopen("/proc/net/dev", "r");
      /* skip column header strings in net/dev */
      fgets(buf, sizeof(buf), fptr); /* skip */
      fgets(buf, sizeof(buf), fptr); /* skip */
      fgets(buf, sizeof(buf), fptr); /* skip */
      fgets(buf, sizeof(buf), fptr); /* eth0 */
      buf[strlen(buf)-1] = 0; /* over '\n' */
      buf[6] = ' ';           /* over  ':' */
      sscanf(buf, "%*s %ld %*d %*d %*d %*d %ld", &rpkts1, &spkts1);
      /* calc network I/O rates */
      stats.pr = rpkts1 - rpkts0;
      stats.ps = spkts1 - spkts0;
      fclose(fptr);

      /*
      **   Reset values for next iteration
      */
      rpkts0 = rpkts1;
      spkts0 = spkts1;
      usec0 += usec1;
      nsec0 += nsec1;
      ssec0 += ssec1;
      isec0 += isec1;
      rio0 = rio1;
      wio0 = wio1;

      /*
      **   Send statistics at 1-second intervals
      */
      sendto(fd, &stats, sizeof(stats), 0,
         (struct sockaddr *) &soc, sizeof(soc));
      sleep(1);
   }
}
