#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <netdb.h>
#include <stdio.h>
#include "msi.h"

main(void)
/*
**   slave.c - inetd service activated by a UDP datagram to port 5000
**
**   inetd configuration (as root - man inetd, or info inetd):
**     1. /etc/services:   slaves 5000/udp     # Master-Slave Interface
**     2. /etc/inetd.conf: slaves dgram udp wait chief /home/chief/slave
**     3. Reconfiguration: killall -HUP inetd
**
**   xinetd configuration (as root - man xinetd.conf, or info xinetd):
**     1. Install xinetd package if necessary
**     2. /etc/xinetd.d/rcatd:
**           service rcatd
**           {
**              port        = 5000
**              socket_type = dgram
**              protocol    = udp
**              wait        = yes
**              user        = chief
**              server      = /home/chief/bin/slave
**              only_from   = 10.0.0.0
**              disable     = no
**           }
**     3. Reconfiguration: /etc/rc.d/init.d/xinetd restart
**
**   In either case:
**     4. Start the Master process first, on any of the cluster's nodes
**        and it will send a broadcast UDP datagram to port 5000, starting
**        a Slave process on every node other than where the Master runs.
*/
{
   int len=sizeof(struct sockaddr_in);
   struct sockaddr_in registers;
   struct sockaddr_in slavelink;
   struct hostent *hp;
#ifdef TIMING
 struct timeval workval;
#endif
   FILE *fp;                      /* query text file pointer */
   int pid;                       /* performance task's pid */
   int line;                      /* requested line number */
   int r_fd;
   int s_fd;
   int opt=1;
   int curr=999;                  /* init current line number */
   char file[80];
   char host[32], *p;
   char buf[QUERYSIZE];

#ifdef TIMING
 /* start remote processing timer */
 gettimeofday(&packet.remote, NULL);
#endif

   /*
   **   Get MASTER's hostname
   */
   read(0, buf, sizeof(buf));
   sscanf(buf, "%s", host);

   /*
   **   Register with the MASTER task
   */
   r_fd = socket(AF_INET, SOCK_DGRAM, 0);
   memset((char *) &registers, 0, len);
   hp = gethostbyname(host); /* get MASTER IP Address */
   memcpy(&registers.sin_addr, hp->h_addr, hp->h_length);
   registers.sin_port = htons(REGISTERS);
   registers.sin_family = AF_INET;
   /* check the local hostname */
   gethostname(buf, sizeof(buf));
   if((p = strchr(buf, '.')) != NULL)
      *p = 0; /* trunc .domainname */
   if(strcmp(host, buf) == 0)
      exit(0); /* no local slaves! */
   sendto(r_fd, buf, strlen(buf)+1, 0,
     (struct sockaddr *) &registers, len);
   close(r_fd);

   /* start "perform" */
   if((pid=fork()) == 0)
   {
      execl("/home/chief/bin/perform", "perform", "cardinal", NULL);
      exit(-1); /* a successful execl never returns to the caller */
   }

   /*
   **   Setup a socket to receive queries
   **   from one of the MASTER's subtasks
   */
   s_fd = socket(AF_INET, SOCK_DGRAM, 0);
   memset((char *) &slavelink, 0, len);
   slavelink.sin_family = AF_INET;
   slavelink.sin_port = htons(SLAVEPORT);
   slavelink.sin_addr.s_addr = htonl(INADDR_ANY);
   bind(s_fd, (struct sockaddr *) &slavelink, len);

   /*
   **   Respond to each query with
   **   the requested line of text
   */
   while(1)
   {
      /* process all incoming queries from Master */
      recvfrom(s_fd, &packet, sizeof(packet),
        0, (struct sockaddr *) &slavelink, &len);

      /*
      **   Get file name & line number from query
      */
#ifdef TIMING
 gettimeofday(&packet.phase4, NULL); /* start Ph4 4 */
#endif
      sscanf(packet.buffer, "%s %d", file, &line);
      if(strncmp(file, "Terminate", 9) == 0)
         break;                                 /* EXIT requested! */
#ifdef TIMING
 gettimeofday(&workval, NULL); /* end Ph4 */
 workval.tv_sec -= packet.phase4.tv_sec;
 workval.tv_usec -= packet.phase4.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 packet.phase4.tv_sec = workval.tv_sec;
 packet.phase4.tv_usec = workval.tv_usec;
 gettimeofday(&packet.phase5, NULL); /* start Ph5 */
#endif

      /*
      **   Extract the requested line of text
      */
      if(line > curr) /* false first time */
      {
         line -= curr; /* adjust to offset */
         curr += line; /* save for next time */
      }
      else
      {
         fclose(fp); /* ignored first time */
         fp = fopen(file, "r");
         curr = line; /* save for next time */
      }
      while(line--)
         fgets(packet.buffer, sizeof(packet.buffer), fp);

#ifdef TIMING
 gettimeofday(&workval, NULL); /* end Ph5 */
 workval.tv_sec -= packet.phase5.tv_sec;
 workval.tv_usec -= packet.phase5.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 packet.phase5.tv_sec = workval.tv_sec;
 packet.phase5.tv_usec = workval.tv_usec;
 gettimeofday(&packet.phase6, NULL); /* start Ph6 */
#endif

      /*
      **   Respond back to Master
      */
      packet.buffer[strlen(packet.buffer)-1] = '\0';
#ifdef TIMING
 gettimeofday(&workval, NULL); /* end Ph6 */
 workval.tv_sec -= packet.phase6.tv_sec;
 workval.tv_usec -= packet.phase6.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 packet.phase6.tv_sec = workval.tv_sec;
 packet.phase6.tv_usec = workval.tv_usec;
#endif
      sendto(s_fd, &packet, sizeof(packet), 0,
        (struct sockaddr *) &slavelink, len);
   } /* while(1) */

   /*
   **   Cleanup and exit
   */
   kill(pid, 9); /* stop performance task */
   waitpid(pid);
   close(s_fd);
   fclose(fp);

#ifdef TIMING
 gettimeofday(&workval, NULL); /* end remote */
 workval.tv_sec -= packet.remote.tv_sec;
 workval.tv_usec -= packet.remote.tv_usec;
 if(workval.tv_usec < 0L)
 {
    workval.tv_usec += 1000000L;
    workval.tv_sec -= 1;
 }
 packet.remote.tv_sec = workval.tv_sec;
 packet.remote.tv_usec = workval.tv_usec;
#endif
   return(0);
}
